package polyglot.ext.polyj.ast;

import polyglot.ext.polyj.extension.*;
import polyglot.ext.polyj.types.*;
import polyglot.ext.polyj.util.*;
import polyglot.ext.polyj.visit.*;

import  polyglot.ext.jl.ast.*;

import polyglot.ast.*;
import polyglot.types.*;
import polyglot.types.Package;
import polyglot.util.*;
import polyglot.visit.*;

import java.util.*;

/**
 * A <code>PolyJField</code> is an immutable representation of a
 * Java field access.  It consists of field name and may also have either a
 * <code>Type</code> or an <code>Expr</code> containing the field being
 * accessed.
 **/
public class PolyJField_c extends Field_c implements PolyJField {

    /**
     * Creates a new <code>PolyJField</code>.
     *
     * @pre <code>target</code> is either a <code>TypeNode</code> or an
     * <code>Expr</code>.
     */

    public PolyJField_c(Position pos, Receiver target, String name)
    {
        super(pos, target, name);
    }

    /**
     * Static fields in polymorphic types are translated to Fetch
     * method calls, casts may need inserted for fields with
     * a polymorphic or parameter type, and some fields
     * (default and length) have special translations.
     */
    public void translate(CodeWriter w, Translator tr) {
        new FieldTranslator(w, (PolyJTranslator) tr).translate();
    }

    class FieldTranslator {
        CodeWriter w;
        Context c;
        PolyJTypeSystem ts;
        TranslationTable tt;
        ReferenceType container;
        ClassType curClass;
        ClassType scopedClass;
        PolyJTranslator tr;
        ClassType uninst_container;
        FieldInstance origFi;
        Type declType;
        Type knownType;
        boolean isStatic; // Is the field static?
        boolean declAsParam; // Is the field declared as a parameter?
        boolean instAsPrim; // Is the field parametric and inst. on a primitive?
        boolean knownAsParam; // Is the field being used at a type parameter?
        boolean declAsParamArray; // Is the field declared as a parametric array?
        boolean isLValue; // Is the field being used as an lvalue?
        
        FieldTranslator(CodeWriter w, PolyJTranslator tr) {
            this.w = w;
            this.tr = tr;
            c = tr.context();
            ts = (PolyJTypeSystem) c.typeSystem();
            tt = ts.translationTable();
            container = fi.container();
            curClass = c.currentClass();
            scopedClass = c.currentClassScope();
        }
        
        void translate() {
            boolean containerNeedsWhereObj = ts.needsWhereObject(container);
            boolean defaultField = container instanceof ParamType 
                && fi.name().equals("default");

            if (containerNeedsWhereObj) {
                translateParametric();
            } else if (defaultField) {
                w.write("(("+ PolyJTranslator.adapterInterfaceFullName(curClass)
                    + ")" + tr.whereObjExpr() + ")." 
                    + ts.DefaultBinding((ParamType) container, curClass).name() 
                    + "()");
            } else if (container.isArray()) {
                translateArray();
            } else {
                translateNormal();
            }
        }
    
    
        /**
         * Translate a field that is a member of a parametric class.
         */
        protected void translateParametric() 
        {
            uninst_container = ts.uninstantiate(container.toClass());
            origFi = ts.originalField(uninst_container, fi);
            declType = origFi.type();
            knownType = fi.type();
            isStatic = fi.flags().isStatic();
            knownAsParam = knownType instanceof ParamType;
            
            // If the field's known type is a parameter, go ahead
            // and finish the translation with translateFieldExpr().  
            // If not, some casts may be necessary,
            // which translateNonParam() will add.           
            if (knownAsParam) {
                translateFieldExpr();
            } else {
                translateNonParam();
            }
        }

        /**
         * Translate an access of a field whose known type is not a type
         * parameter, and that is contained by a parametric class.
         */
        protected void translateNonParam() 
        {
            declAsParam = declType instanceof ParamType;
            instAsPrim = knownType.isPrimitive() && declAsParam;
            declAsParamArray = declType.isArray() &&
                declType.toArray().ultimateBase() instanceof ParamType;
            isLValue = ((PolyJExt) PolyJField_c.this.ext()).isLValue();

            if (declAsParam || declAsParamArray && !isLValue) {
                String cast = "";
                if (instAsPrim) {
                    cast = ts.wrapperTypeString(knownType.toPrimitive());
                } else {
                    cast = knownType.translate(c);
                }
                w.write("((" + cast + ")");
                translateFieldExpr();
                w.write(")");
                if (instAsPrim) {
                    w.write(".value");
                }
            } else {
                translateFieldExpr();
            }
        }

        /**
         * Finish the translation of a field by producing the
         * expression that will access it.
         */
        protected void translateFieldExpr() {
            if (isStatic) {
                String whereExpr = null;
                if (ts.isFullyInst(container.toClass())) {
                    whereExpr = ts.InstantiationBinding(container.toClass())
                        .getAdapterClassFullName() + ".getInstance()";
                } else {
                    if (container.equals(curClass)) {
                        whereExpr = "(("
                            + PolyJTranslator.adapterInterfaceFullName(curClass) 
                            + ")" + tr.whereObjExpr() + ")";
                    } else {
                        AdapterBinding ab = ts.AdapterBinding(
                            (PolyJClassType) scopedClass, container.toClass());
                        whereExpr = "(("
                            + PolyJTranslator.adapterInterfaceFullName(curClass) 
                            + ")" + tr.whereObjExpr() + ")." + ab.name()+"()";
                    }
                }
                translateStaticGet(origFi, uninst_container, whereExpr);
            } else {
                translateNormal();
            }        
        }

        /**
         * Translate the use of A.length, where A:array[X].
         * If X is a ParamType, this requires a special translation.
         */
        protected void translateArray() {
            ArrayType at = container.toArray();
            Context c = tr.context();
            PolyJTypeSystem ts = (PolyJTypeSystem) c.typeSystem();
            TranslationTable tt = ts.translationTable();
            PolyJTranslator pjtr = (PolyJTranslator) tr;

            //arrays should only have one field:  length
            if (!fi.name().equals("length")) {
                throw new InternalCompilerError(position(),
                    fi.name() + " is not a legal field for an array");
            }

            if (at.base() instanceof ParamType) {
                ParamType elementType = (ParamType)at.base();
                ClassType curClass = c.currentClass();
                ArrayAccessBinding binding = ts.ArrayAccessBinding(
                    elementType, curClass);
                w.write("((" 
                    + PolyJTranslator.adapterInterfaceFullName(curClass) 
                    + ")" + pjtr.whereObjExpr() + ")." + binding.length() + "(");
                w.begin(0);
                printSubExpr((Expr) target, w, tr);
                w.write(")");
                w.end();
            } else {
                translateNormal();
            }
        }

        /**
         * Translate the access of a static field of a polymorphic class,
         * which requires inserting a fetch method call.
         */
        protected void translateStaticGet(FieldInstance fi, 
            ClassType enclosingPolyType, String adapterClassName) 
        {
            StaticFieldBinding fieldBinding =
                ts.StaticFieldBinding(enclosingPolyType, fi);

            Package pack = enclosingPolyType.package_();
            String packageName = "";
            if (pack != null) {
                packageName = pack.fullName();
            }

            w.write(adapterClassName+"."+fieldBinding.get()+"(");
            if (!fi.flags().isPublic()) {
                if (packageName == "") {
                    w.write("$token.the_instance");
                } else {
                    w.write(packageName + ".$token.the_instance");
                }
            }
            w.write(")" );
        }

        /**
         * Translate the target of the field.  This behaves
         * slightly differently than the superclass in that
         * it omits the target if it isn't needed.
         */
        protected void translateTarget() {
            Context c = tr.context();

            // Translate whatever is affixed onto the left of the dot
            if (target != null) {
                if (target instanceof Expr) {
                    printSubExpr((Expr) target, w, tr);
                    w.write( ".");
                } else if (target instanceof TypeNode) {
                    if (((TypeNode) target).type() != c.currentClass()
                        || name.equals( "class")) {
                        target.del().translate(w, tr);
                        w.write( ".");
                    }
                } // Expr and TypeNode are the only two possibilities
            }
        }

        /**
         * Do a normal translation, with just a small change
         * for the target (see translateTarget()).
         */
        protected void translateNormal() {
            translateTarget();
            w.write(name);
        }
    }
        
    public void dump(CodeWriter w)
    {
        w.write( "POLYJ FIELD ACCESS");
        w.write( " < " + name + " > ");
    }
}
