package polyglot.ext.polyj.parse;

import polyglot.ast.*;
import polyglot.visit.*;
import polyglot.types.*;
import polyglot.util.*;
import polyglot.parse.*;

import polyglot.ext.polyj.ast.*;

import java.util.*;

/**
 * A <code>Name</code> represents a <code>Parse</code> of the form "n | P.n".  
 * This could either be a field access or a type.  Both could be preceded by
 * ambiguous qualifiers.
 */  
public class Name extends Parse {
    String name;
    Parse prefix;

    public Name(BaseParser parser, Position pos, String name) throws Exception {
        this(parser, pos, null, name);
    }

    public Name(BaseParser parser, Position pos, Parse prefix, String name) throws Exception {
        super(parser, pos);
        this.prefix = prefix;
        this.name = name;

        if (prefix instanceof Array) parser.die("Illegal name: preceded by array type.", pos);
    }

    public boolean isIdentifier() {
        return prefix == null;
    }
    
    public boolean isQualified() {
        return prefix != null;
    }
    
    public Call toCall(List args) throws Exception {
        Receiver receiver = null;
        if (prefix != null) {
            receiver = prefix.toReceiver();
        }
        
        return nf.Call(pos, receiver, name, args);
    }
    
    public Expr toExpr() throws Exception {
        if (prefix == null) {
            return nf.AmbExpr(pos, name);
        }
        
        return nf.Field(pos, prefix.toReceiver(), name);    
    }

    public AmbReceiver toAmbReceiver() throws Exception {
        if (prefix == null) {
            return nf.AmbReceiver(pos, name);
        }

        return nf.AmbReceiver(pos, prefix.toPrefix(), name);
    }
    
    public Receiver toReceiver() throws Exception {
        return toAmbReceiver();
    }

    public Prefix toPrefix() throws Exception {
        if (prefix == null) {
            return nf.AmbPrefix(pos, name);
        }

        return nf.AmbPrefix(pos, prefix.toPrefix(), name);
    }

    public QualifierNode toQualifier() throws Exception {
        if (prefix == null) {
            return nf.AmbQualifierNode(pos, name);
        }

        return nf.AmbQualifierNode(pos, prefix.toQualifier(), name);
    }

    public PackageNode toPackage() throws Exception {
        return nf.PackageNode(pos, ts.packageForName(toName()));
    }

    public TypeNode toType() throws Exception {
        if (prefix == null) {
            return nf.AmbTypeNode(pos, name);
        }
        
        return nf.AmbTypeNode(pos, prefix.toQualifier(), name);
    }

    public TypeNode toClassType() throws Exception { 
        return toType(); 
    }

    public String toIdentifier() throws Exception {
        if (prefix != null) {
            parser.die("Illegal identifier: preceded by qualified name", pos);
        }

        return name;
    }

    public String toName() throws Exception {
        if (prefix == null) {
            return name;
        }

        return prefix.toName() + "." + name;
    }
    
    public String toString() {
        if (prefix == null) {
            return name;
        }
        
        return prefix + "." + name;
    }
    
    public Parse prepend(Parse base) throws Exception {
        return new InstOrAccess(parser, pos, base, this);
    }
    
    public Receiver toParam() throws Exception {
        return toReceiver();    
    }
}
